==> libixion-0.5.0.tar.bz2: libixion-0.5.0/src/ixion_test.cpp <==
/*************************************************************************
 *
 * Copyright (c) 2011-2012 Kohei Yoshida
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 ************************************************************************/

#include "ixion/formula_name_resolver.hpp"
#include "ixion/address.hpp"
#include "ixion/formula.hpp"
#include "ixion/model_context.hpp"
#include "ixion/cell.hpp"
#include "ixion/global.hpp"
#include "ixion/macros.hpp"

#include <iostream>
#include <cassert>
#include <string>
#include <cstring>
#include <sstream>

using namespace std;
using namespace ixion;

namespace {

void test_size()
{
    cout << "test size" << endl;
    cout << "int: " << sizeof(int) << endl;
    cout << "long: " << sizeof(long) << endl;
    cout << "double: " << sizeof(double) << endl;
    cout << "size_t: " << sizeof(size_t) << endl;
    cout << "celltype_t: " << sizeof(celltype_t) << endl;
    cout << "formula_cell: " << sizeof(formula_cell) << endl;
    cout << "formula_tokens_t: " << sizeof(formula_tokens_t) << endl;
}

void test_string_to_double()
{
    cout << "test string to double" << endl;
    struct { const char* s; double v; } tests[] = {
        { "12", 12.0 },
        { "0", 0.0 },
        { "1.3", 1.3 },
        { "1234.00983", 1234.00983 },
        { "-123.3", -123.3 }
    };

    size_t n = sizeof(tests) / sizeof(tests[0]);
    for (size_t i = 0; i < n; ++i)
    {
        double v = global::to_double(tests[i].s, strlen(tests[i].s));
        assert(v == tests[i].v);
    }
}

void test_name_resolver()
{
    cout << "test name resolver" << endl;

    model_context cxt;
    cxt.append_sheet(IXION_ASCII("One"));
    cxt.append_sheet(IXION_ASCII("Two"));
    cxt.append_sheet(IXION_ASCII("Three"));
    cxt.append_sheet(IXION_ASCII("A B C")); // name with space
    formula_name_resolver_a1 resolver(&cxt);

    // Parse single cell addresses.
    struct {
        const char* name; bool sheet_name;
    } names[] = {
        { "A1", false },
        { "$A1", false },
        { "A$1", false },
        { "$A$1", false },
        { "Z1", false },
        { "AA23", false },
        { "AB23", false },
        { "$AB23", false },
        { "AB$23", false },
        { "$AB$23", false },
        { "BA1", false },
        { "AAA2", false },
        { "ABA1", false },
        { "BAA1", false },
        { "XFD1048576", false },
        { "One!A1", true },
        { "One!XFD1048576", true },
        { "Two!B10", true },
        { "Two!$B10", true },
        { "Two!B$10", true },
        { "Two!$B$10", true },
        { "Three!CFD234", true },
        { "'A B C'!Z12", true },
        { 0, false }
    };

    for (size_t i = 0; names[i].name; ++i)
    {
        const char* p = names[i].name;
        string name_a1(p);
        formula_name_type res = resolver.resolve(&name_a1[0], name_a1.size(), abs_address_t());
        if (res.type != formula_name_type::cell_reference)
        {
            cerr << "failed to resolve cell address: " << name_a1 << endl;
            assert(false);
        }

        address_t addr;
        addr.sheet = res.address.sheet;
        addr.abs_sheet = res.address.abs_sheet;
        addr.row = res.address.row;
        addr.abs_row = res.address.abs_row;
        addr.column = res.address.col;
        addr.abs_column = res.address.abs_col;

        string test_name = resolver.get_name(addr, abs_address_t(), names[i].sheet_name);
        if (name_a1 != test_name)
        {
            cerr << "failed to compile name from address: (name expected: " << name_a1 << "; actual name created: " << test_name << ")" << endl;
            assert(false);
        }
    }

    // Parse range addresses.
    struct {
        const char* name; sheet_t sheet1; row_t row1; col_t col1; sheet_t sheet2; row_t row2; col_t col2;
    } range_tests[] = {
        { "A1:B2", 0, 0, 0, 0, 1, 1 },
        { "D10:G24", 0, 9, 3, 0, 23, 6 },
        { "One!C1:Z400", 0, 0, 2, 0, 399, 25 },
        { "Two!C1:Z400", 1, 0, 2, 1, 399, 25 },
        { "Three!C1:Z400", 2, 0, 2, 2, 399, 25 },
        { 0, 0, 0, 0, 0, 0, 0 }
    };

    for (size_t i = 0; range_tests[i].name; ++i)
    {
        string name_a1(range_tests[i].name);
        formula_name_type res = resolver.resolve(&name_a1[0], name_a1.size(), abs_address_t());
        assert(res.type == formula_name_type::range_reference);
        assert(res.range.first.sheet == range_tests[i].sheet1);
        assert(res.range.first.row == range_tests[i].row1);
        assert(res.range.first.col == range_tests[i].col1);
        assert(res.range.last.sheet == range_tests[i].sheet2);
        assert(res.range.last.row == range_tests[i].row2);
        assert(res.range.last.col == range_tests[i].col2);
    }

    formula_name_type res = resolver.resolve("B1", 2, abs_address_t(0,1,1));
    assert(res.type == formula_name_type::cell_reference);
    assert(res.address.sheet == 0);
    assert(res.address.row == -1);
    assert(res.address.col == 0);

    res = resolver.resolve("B2:B4", 5, abs_address_t(0,0,3));
    assert(res.type == formula_name_type::range_reference);
    assert(res.range.first.sheet == 0);
    assert(res.range.first.row == 1);
    assert(res.range.first.col == -2);
    assert(res.range.last.sheet == 0);
    assert(res.range.last.row == 3);
    assert(res.range.last.col == -2);

    // Parse name without row index.
    struct {
        const char* name; formula_name_type::name_type type;
    } name_tests[] = {
        { "H:H", formula_name_type::range_reference },
        { "ABC", formula_name_type::named_expression },
        { "H", formula_name_type::named_expression },
        { "MAX", formula_name_type::function },
        { 0, formula_name_type::invalid }
    };

    for (size_t i = 0; name_tests[i].name; ++i)
    {
        string name_a1(name_tests[i].name);
        formula_name_type res = resolver.resolve(&name_a1[0], name_a1.size(), abs_address_t());
        assert(res.type == name_tests[i].type);
    }
}

void test_address()
{
    cout << "test address" << endl;
    address_t addr(-1, 0, 0, false, false, false);
    abs_address_t pos(1, 0, 0);
    abs_address_t abs_addr = addr.to_abs(pos);
    assert(abs_addr.sheet == 0 && abs_addr.row == 0 && abs_addr.column == 0);

    // Default constructor makes valid address.
    assert(abs_address_t().valid());
    assert(abs_range_t().valid());

    // These are invalid addresses.
    assert(!abs_address_t(abs_address_t::invalid).valid());
    assert(!abs_range_t(abs_range_t::invalid).valid());
}

bool check_formula_expression(model_context& cxt, const char* p)
{
    size_t n = strlen(p);
    cout << "testing formula expression '" << p << "'" << endl;
    formula_tokens_t tokens;
    parse_formula_string(cxt, abs_address_t(), p, n, tokens);
    std::string str;
    print_formula_tokens(cxt, abs_address_t(), tokens, str);
    int res = strcmp(p, str.c_str());
    if (res)
        cout << "formula expressions differ: '" << p << "' (before) -> '" << str << "' (after)" << endl;

    return res == 0;
}

/**
 * Make sure the public API works as advertized.
 */
void test_parse_and_print_expressions()
{
    cout << "test public formula api" << endl;
    const char* exps[] = {
        "1/3*1.4",
        "2.3*(1+2)/(34*(3-2))",
        "SUM(1,2,3)",
        "A1",
        "B10",
        "XFD1048576",
        "C10:D20",
        "A1:XFD1048576",
        "H:H",
        "B:D",
        "AB:AD",
        "2:2",
        "3:5",
        "34:36"
    };
    size_t num_exps = sizeof(exps) / sizeof(exps[0]);
    model_context cxt;
    for (size_t i = 0; i < num_exps; ++i)
    {
        bool result = check_formula_expression(cxt, exps[i]);
        assert(result);
    }
}

/**
 * Function name must be resolved case-insensitively.
 */
void test_function_name_resolution()
{
    cout << "test function name resolution" << endl;

    const char* valid_names[] = {
        "SUM", "sum", "Sum", "Average", "max", "min"
    };

    const char* invalid_names[] = {
        "suma", "foo", "", "su", "maxx", "minmin"
    };

    model_context cxt;
    const formula_name_resolver& resolver = cxt.get_name_resolver();
    size_t n = IXION_N_ELEMENTS(valid_names);
    for (size_t i = 0; i < n; ++i)
    {
        const char* name = valid_names[i];
        cout << "valid name: " << name << endl;
        formula_name_type t = resolver.resolve(name, strlen(name), abs_address_t());
        assert(t.type == formula_name_type::function);
    }

    n = IXION_N_ELEMENTS(invalid_names);
    for (size_t i = 0; i < n; ++i)
    {
        const char* name = invalid_names[i];
        cout << "invalid name: " << name << endl;
        formula_name_type t = resolver.resolve(name, strlen(name), abs_address_t());
        assert(t.type != formula_name_type::function);
    }
}

formula_cell* insert_formula(model_context& cxt, const abs_address_t& pos, const char* exp)
{
    cxt.set_formula_cell(pos, exp, strlen(exp));
    register_formula_cell(cxt, pos);
    formula_cell* p = cxt.get_formula_cell(pos);
    assert(p);
    return p;
}

void test_model_context_storage()
{
    cout << "test model context storage" << endl;
    {
        model_context cxt;
        cxt.append_sheet(IXION_ASCII("test"));
        cxt.set_session_handler(NULL);

        // Test storage of numeric values.
        double val = 0.1;
        for (col_t col = 0; col < 3; ++col)
        {
            for (row_t row = 0; row < 3; ++row)
            {
                abs_address_t pos(0, row, col);
                cxt.set_numeric_cell(pos, val);
                double test = cxt.get_numeric_value(pos);
                assert(test == val);
                val += 0.2;
            }
        }

        // Test formula cells.
        abs_address_t pos(0,3,0);
        const char* exp = "SUM(1,2,3)";
        cxt.set_formula_cell(pos, exp, strlen(exp));
        formula_cell* p = cxt.get_formula_cell(pos);
    }

    {
        model_context cxt;
        cxt.append_sheet(IXION_ASCII("test"));
        cxt.set_session_handler(NULL);
        string exp = "1";
        cxt.set_formula_cell(abs_address_t(0,0,0), &exp[0], exp.size());
        cxt.set_formula_cell(abs_address_t(0,2,0), &exp[0], exp.size());
        cxt.set_formula_cell(abs_address_t(0,1,0), &exp[0], exp.size());
    }

    {
        // Test data area.
        model_context cxt(10, 10);
        cxt.append_sheet(IXION_ASCII("test"));

        abs_range_t area = cxt.get_data_range(0);
        assert(!area.valid());

        cxt.set_numeric_cell(abs_address_t(0, 6, 5), 1.1);
        area = cxt.get_data_range(0);
        assert(area.first == area.last);
        assert(area.first.sheet == 0);
        assert(area.first.row == 6);
        assert(area.first.column == 5);

        cxt.set_numeric_cell(abs_address_t(0, 2, 3), 1.1);
        area = cxt.get_data_range(0);
        assert(area.first.sheet == 0);
        assert(area.first.row == 2);
        assert(area.first.column == 3);
        assert(area.last.sheet == 0);
        assert(area.last.row == 6);
        assert(area.last.column == 5);

        cxt.set_numeric_cell(abs_address_t(0, 7, 1), 1.1);
        area = cxt.get_data_range(0);
        assert(area.first.sheet == 0);
        assert(area.first.row == 2);
        assert(area.first.column == 1);
        assert(area.last.sheet == 0);
        assert(area.last.row == 7);
        assert(area.last.column == 5);

        // This shouldn't change the data range.
        cxt.set_numeric_cell(abs_address_t(0, 5, 5), 1.1);
        abs_range_t test = cxt.get_data_range(0);
        assert(test == area);
    }
}

void test_volatile_function()
{
    cout << "test volatile function" << endl;

    model_context cxt;
    cxt.append_sheet(IXION_ASCII("test"));
    cxt.set_session_handler(NULL);

    dirty_formula_cells_t dirty_cells;
    modified_cells_t dirty_addrs;

    // Set values into A1:A3.
    cxt.set_numeric_cell(abs_address_t(0,0,0), 1.0);
    cxt.set_numeric_cell(abs_address_t(0,1,0), 2.0);
    cxt.set_numeric_cell(abs_address_t(0,2,0), 3.0);

    // Set formula in A4 that references A1:A3.
    formula_cell* p = insert_formula(cxt, abs_address_t(0,3,0), "SUM(A1:A3)");
    assert(p);
    dirty_cells.insert(abs_address_t(0,3,0));

    // Initial full calculation.
    calculate_cells(cxt, dirty_cells, 0);

    double val = cxt.get_numeric_value(abs_address_t(0,3,0));
    assert(val == 6);

    // Modify the value of A2.  This should flag A4 dirty.
    cxt.set_numeric_cell(abs_address_t(0,1,0), 10.0);
    dirty_cells.clear();
    dirty_addrs.push_back(abs_address_t(0,1,0));
    get_all_dirty_cells(cxt, dirty_addrs, dirty_cells);
    assert(dirty_cells.size() == 1);

    // Partial recalculation.
    calculate_cells(cxt, dirty_cells, 0);
    val = cxt.get_numeric_value(abs_address_t(0, 3, 0));
    assert(val == 14);

    // Insert a volatile cell into B1.  At this point B1 should be the only dirty cell.
    dirty_cells.clear();
    dirty_addrs.clear();
    p = insert_formula(cxt, abs_address_t(0,0,1), "NOW()");
    assert(p);
    dirty_cells.insert(abs_address_t(0,0,1));
    dirty_addrs.push_back(abs_address_t(0,0,1));
    get_all_dirty_cells(cxt, dirty_addrs, dirty_cells);
    assert(dirty_cells.size() == 1);

    // Partial recalc again.
    calculate_cells(cxt, dirty_cells, 0);
    double t1 = cxt.get_numeric_value(abs_address_t(0,0,1));

    // Pause for 0.2 second.
    global::sleep(200);

    // No modification, but B1 should still be flagged dirty.
    dirty_cells.clear();
    dirty_addrs.clear();
    get_all_dirty_cells(cxt, dirty_addrs, dirty_cells);
    assert(dirty_cells.size() == 1);
    calculate_cells(cxt, dirty_cells, 0);
    double t2 = cxt.get_numeric_value(abs_address_t(0,0,1));
    double delta = (t2-t1)*24*60*60;
    cout << "delta = " << delta << endl;

    // The delta should be close to 0.2.  It may be a little larger depending
    // on the CPU speed.
    assert(0.2 <= delta && delta <= 0.3);
}

}

int main()
{
    test_size();
    test_string_to_double();
    test_name_resolver();
    test_address();
    test_parse_and_print_expressions();
    test_function_name_resolution();
    test_model_context_storage();
    test_volatile_function();
    return EXIT_SUCCESS;
}

==> libixion_0.7.0.orig.tar.bz2: libixion-0.7.0/src/ixion_test.cpp <==
/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "ixion/formula_name_resolver.hpp"
#include "ixion/address.hpp"
#include "ixion/formula.hpp"
#include "ixion/model_context.hpp"
#include "ixion/cell.hpp"
#include "ixion/global.hpp"
#include "ixion/macros.hpp"

#include <iostream>
#include <cassert>
#include <string>
#include <cstring>
#include <sstream>

using namespace std;
using namespace ixion;

namespace {

void test_size()
{
    cout << "test size" << endl;
    cout << "int: " << sizeof(int) << endl;
    cout << "long: " << sizeof(long) << endl;
    cout << "double: " << sizeof(double) << endl;
    cout << "size_t: " << sizeof(size_t) << endl;
    cout << "celltype_t: " << sizeof(celltype_t) << endl;
    cout << "formula_cell: " << sizeof(formula_cell) << endl;
    cout << "formula_tokens_t: " << sizeof(formula_tokens_t) << endl;
}

void test_string_to_double()
{
    cout << "test string to double" << endl;
    struct { const char* s; double v; } tests[] = {
        { "12", 12.0 },
        { "0", 0.0 },
        { "1.3", 1.3 },
        { "1234.00983", 1234.00983 },
        { "-123.3", -123.3 }
    };

    size_t n = sizeof(tests) / sizeof(tests[0]);
    for (size_t i = 0; i < n; ++i)
    {
        double v = global::to_double(tests[i].s, strlen(tests[i].s));
        assert(v == tests[i].v);
    }
}

void test_name_resolver()
{
    cout << "test name resolver" << endl;

    model_context cxt;
    cxt.append_sheet(IXION_ASCII("One"), 1048576, 1024);
    cxt.append_sheet(IXION_ASCII("Two"), 1048576, 1024);
    cxt.append_sheet(IXION_ASCII("Three"), 1048576, 1024);
    cxt.append_sheet(IXION_ASCII("A B C"), 1048576, 1024); // name with space
    formula_name_resolver_a1 resolver(&cxt);

    // Parse single cell addresses.
    struct {
        const char* name; bool sheet_name;
    } names[] = {
        { "A1", false },
        { "$A1", false },
        { "A$1", false },
        { "$A$1", false },
        { "Z1", false },
        { "AA23", false },
        { "AB23", false },
        { "$AB23", false },
        { "AB$23", false },
        { "$AB$23", false },
        { "BA1", false },
        { "AAA2", false },
        { "ABA1", false },
        { "BAA1", false },
        { "XFD1048576", false },
        { "One!A1", true },
        { "One!XFD1048576", true },
        { "Two!B10", true },
        { "Two!$B10", true },
        { "Two!B$10", true },
        { "Two!$B$10", true },
        { "Three!CFD234", true },
        { "'A B C'!Z12", true },
        { 0, false }
    };

    for (size_t i = 0; names[i].name; ++i)
    {
        const char* p = names[i].name;
        string name_a1(p);
        formula_name_type res = resolver.resolve(&name_a1[0], name_a1.size(), abs_address_t());
        if (res.type != formula_name_type::cell_reference)
        {
            cerr << "failed to resolve cell address: " << name_a1 << endl;
            assert(false);
        }

        address_t addr;
        addr.sheet = res.address.sheet;
        addr.abs_sheet = res.address.abs_sheet;
        addr.row = res.address.row;
        addr.abs_row = res.address.abs_row;
        addr.column = res.address.col;
        addr.abs_column = res.address.abs_col;

        string test_name = resolver.get_name(addr, abs_address_t(), names[i].sheet_name);
        if (name_a1 != test_name)
        {
            cerr << "failed to compile name from address: (name expected: " << name_a1 << "; actual name created: " << test_name << ")" << endl;
            assert(false);
        }
    }

    // Parse range addresses.
    struct {
        const char* name; sheet_t sheet1; row_t row1; col_t col1; sheet_t sheet2; row_t row2; col_t col2;
    } range_tests[] = {
        { "A1:B2", 0, 0, 0, 0, 1, 1 },
        { "D10:G24", 0, 9, 3, 0, 23, 6 },
        { "One!C1:Z400", 0, 0, 2, 0, 399, 25 },
        { "Two!C1:Z400", 1, 0, 2, 1, 399, 25 },
        { "Three!C1:Z400", 2, 0, 2, 2, 399, 25 },
        { 0, 0, 0, 0, 0, 0, 0 }
    };

    for (size_t i = 0; range_tests[i].name; ++i)
    {
        string name_a1(range_tests[i].name);
        formula_name_type res = resolver.resolve(&name_a1[0], name_a1.size(), abs_address_t());
        assert(res.type == formula_name_type::range_reference);
        assert(res.range.first.sheet == range_tests[i].sheet1);
        assert(res.range.first.row == range_tests[i].row1);
        assert(res.range.first.col == range_tests[i].col1);
        assert(res.range.last.sheet == range_tests[i].sheet2);
        assert(res.range.last.row == range_tests[i].row2);
        assert(res.range.last.col == range_tests[i].col2);
    }

    formula_name_type res = resolver.resolve("B1", 2, abs_address_t(0,1,1));
    assert(res.type == formula_name_type::cell_reference);
    assert(res.address.sheet == 0);
    assert(res.address.row == -1);
    assert(res.address.col == 0);

    res = resolver.resolve("B2:B4", 5, abs_address_t(0,0,3));
    assert(res.type == formula_name_type::range_reference);
    assert(res.range.first.sheet == 0);
    assert(res.range.first.row == 1);
    assert(res.range.first.col == -2);
    assert(res.range.last.sheet == 0);
    assert(res.range.last.row == 3);
    assert(res.range.last.col == -2);

    // Parse name without row index.
    struct {
        const char* name; formula_name_type::name_type type;
    } name_tests[] = {
        { "H:H", formula_name_type::range_reference },
        { "ABC", formula_name_type::named_expression },
        { "H", formula_name_type::named_expression },
        { "MAX", formula_name_type::function },
        { 0, formula_name_type::invalid }
    };

    for (size_t i = 0; name_tests[i].name; ++i)
    {
        string name_a1(name_tests[i].name);
        formula_name_type res = resolver.resolve(&name_a1[0], name_a1.size(), abs_address_t());
        assert(res.type == name_tests[i].type);
    }
}

void test_address()
{
    cout << "test address" << endl;
    address_t addr(-1, 0, 0, false, false, false);
    abs_address_t pos(1, 0, 0);
    abs_address_t abs_addr = addr.to_abs(pos);
    assert(abs_addr.sheet == 0 && abs_addr.row == 0 && abs_addr.column == 0);

    // Default constructor makes valid address.
    assert(abs_address_t().valid());
    assert(abs_range_t().valid());

    // These are invalid addresses.
    assert(!abs_address_t(abs_address_t::invalid).valid());
    assert(!abs_range_t(abs_range_t::invalid).valid());
}

bool check_formula_expression(model_context& cxt, const char* p)
{
    size_t n = strlen(p);
    cout << "testing formula expression '" << p << "'" << endl;
    formula_tokens_t tokens;
    parse_formula_string(cxt, abs_address_t(), p, n, tokens);
    std::string str;
    print_formula_tokens(cxt, abs_address_t(), tokens, str);
    int res = strcmp(p, str.c_str());
    if (res)
        cout << "formula expressions differ: '" << p << "' (before) -> '" << str << "' (after)" << endl;

    return res == 0;
}

/**
 * Make sure the public API works as advertized.
 */
void test_parse_and_print_expressions()
{
    cout << "test public formula api" << endl;
    const char* exps[] = {
        "1/3*1.4",
        "2.3*(1+2)/(34*(3-2))",
        "SUM(1,2,3)",
        "A1",
        "B10",
        "XFD1048576",
        "C10:D20",
        "A1:XFD1048576",
        "H:H",
        "B:D",
        "AB:AD",
        "2:2",
        "3:5",
        "34:36"
    };
    size_t num_exps = sizeof(exps) / sizeof(exps[0]);
    model_context cxt;
    for (size_t i = 0; i < num_exps; ++i)
    {
        bool result = check_formula_expression(cxt, exps[i]);
        assert(result);
    }
}

/**
 * Function name must be resolved case-insensitively.
 */
void test_function_name_resolution()
{
    cout << "test function name resolution" << endl;

    const char* valid_names[] = {
        "SUM", "sum", "Sum", "Average", "max", "min"
    };

    const char* invalid_names[] = {
        "suma", "foo", "", "su", "maxx", "minmin"
    };

    model_context cxt;
    const formula_name_resolver& resolver = cxt.get_name_resolver();
    size_t n = IXION_N_ELEMENTS(valid_names);
    for (size_t i = 0; i < n; ++i)
    {
        const char* name = valid_names[i];
        cout << "valid name: " << name << endl;
        formula_name_type t = resolver.resolve(name, strlen(name), abs_address_t());
        assert(t.type == formula_name_type::function);
    }

    n = IXION_N_ELEMENTS(invalid_names);
    for (size_t i = 0; i < n; ++i)
    {
        const char* name = invalid_names[i];
        cout << "invalid name: " << name << endl;
        formula_name_type t = resolver.resolve(name, strlen(name), abs_address_t());
        assert(t.type != formula_name_type::function);
    }
}

formula_cell* insert_formula(model_context& cxt, const abs_address_t& pos, const char* exp)
{
    cxt.set_formula_cell(pos, exp, strlen(exp));
    register_formula_cell(cxt, pos);
    formula_cell* p = cxt.get_formula_cell(pos);
    assert(p);
    return p;
}

void test_model_context_storage()
{
    cout << "test model context storage" << endl;
    {
        model_context cxt;
        cxt.append_sheet(IXION_ASCII("test"), 1048576, 1024);
        cxt.set_session_handler(NULL);

        // Test storage of numeric values.
        double val = 0.1;
        for (col_t col = 0; col < 3; ++col)
        {
            for (row_t row = 0; row < 3; ++row)
            {
                abs_address_t pos(0, row, col);
                cxt.set_numeric_cell(pos, val);
                double test = cxt.get_numeric_value(pos);
                assert(test == val);
                val += 0.2;
            }
        }

        // Test formula cells.
        abs_address_t pos(0,3,0);
        const char* exp = "SUM(1,2,3)";
        cxt.set_formula_cell(pos, exp, strlen(exp));
        formula_cell* p = cxt.get_formula_cell(pos);
        assert(p);
    }

    {
        model_context cxt;
        cxt.append_sheet(IXION_ASCII("test"), 1048576, 1024);
        cxt.set_session_handler(NULL);
        string exp = "1";
        cxt.set_formula_cell(abs_address_t(0,0,0), &exp[0], exp.size());
        cxt.set_formula_cell(abs_address_t(0,2,0), &exp[0], exp.size());
        cxt.set_formula_cell(abs_address_t(0,1,0), &exp[0], exp.size());
    }

    {
        // Test data area.
        model_context cxt;
        cxt.append_sheet(IXION_ASCII("test"), 1048576, 1024);

        abs_range_t area = cxt.get_data_range(0);
        assert(!area.valid());

        cxt.set_numeric_cell(abs_address_t(0, 6, 5), 1.1);
        area = cxt.get_data_range(0);
        assert(area.first == area.last);
        assert(area.first.sheet == 0);
        assert(area.first.row == 6);
        assert(area.first.column == 5);

        cxt.set_numeric_cell(abs_address_t(0, 2, 3), 1.1);
        area = cxt.get_data_range(0);
        assert(area.first.sheet == 0);
        assert(area.first.row == 2);
        assert(area.first.column == 3);
        assert(area.last.sheet == 0);
        assert(area.last.row == 6);
        assert(area.last.column == 5);

        cxt.set_numeric_cell(abs_address_t(0, 7, 1), 1.1);
        area = cxt.get_data_range(0);
        assert(area.first.sheet == 0);
        assert(area.first.row == 2);
        assert(area.first.column == 1);
        assert(area.last.sheet == 0);
        assert(area.last.row == 7);
        assert(area.last.column == 5);

        // This shouldn't change the data range.
        cxt.set_numeric_cell(abs_address_t(0, 5, 5), 1.1);
        abs_range_t test = cxt.get_data_range(0);
        assert(test == area);
    }
}

void test_volatile_function()
{
    cout << "test volatile function" << endl;

    model_context cxt;
    cxt.append_sheet(IXION_ASCII("test"), 1048576, 1024);
    cxt.set_session_handler(NULL);

    dirty_formula_cells_t dirty_cells;
    modified_cells_t dirty_addrs;

    // Set values into A1:A3.
    cxt.set_numeric_cell(abs_address_t(0,0,0), 1.0);
    cxt.set_numeric_cell(abs_address_t(0,1,0), 2.0);
    cxt.set_numeric_cell(abs_address_t(0,2,0), 3.0);

    // Set formula in A4 that references A1:A3.
    formula_cell* p = insert_formula(cxt, abs_address_t(0,3,0), "SUM(A1:A3)");
    assert(p);
    dirty_cells.insert(abs_address_t(0,3,0));

    // Initial full calculation.
    calculate_cells(cxt, dirty_cells, 0);

    double val = cxt.get_numeric_value(abs_address_t(0,3,0));
    assert(val == 6);

    // Modify the value of A2.  This should flag A4 dirty.
    cxt.set_numeric_cell(abs_address_t(0,1,0), 10.0);
    dirty_cells.clear();
    dirty_addrs.push_back(abs_address_t(0,1,0));
    get_all_dirty_cells(cxt, dirty_addrs, dirty_cells);
    assert(dirty_cells.size() == 1);

    // Partial recalculation.
    calculate_cells(cxt, dirty_cells, 0);
    val = cxt.get_numeric_value(abs_address_t(0, 3, 0));
    assert(val == 14);

    // Insert a volatile cell into B1.  At this point B1 should be the only dirty cell.
    dirty_cells.clear();
    dirty_addrs.clear();
    p = insert_formula(cxt, abs_address_t(0,0,1), "NOW()");
    assert(p);
    dirty_cells.insert(abs_address_t(0,0,1));
    dirty_addrs.push_back(abs_address_t(0,0,1));
    get_all_dirty_cells(cxt, dirty_addrs, dirty_cells);
    assert(dirty_cells.size() == 1);

    // Partial recalc again.
    calculate_cells(cxt, dirty_cells, 0);
    double t1 = cxt.get_numeric_value(abs_address_t(0,0,1));

    // Pause for 0.2 second.
    global::sleep(200);

    // No modification, but B1 should still be flagged dirty.
    dirty_cells.clear();
    dirty_addrs.clear();
    get_all_dirty_cells(cxt, dirty_addrs, dirty_cells);
    assert(dirty_cells.size() == 1);
    calculate_cells(cxt, dirty_cells, 0);
    double t2 = cxt.get_numeric_value(abs_address_t(0,0,1));
    double delta = (t2-t1)*24*60*60;
    cout << "delta = " << delta << endl;

    // The delta should be close to 0.2.  It may be a little larger depending
    // on the CPU speed.
    assert(0.2 <= delta && delta <= 0.3);
}

}

int main()
{
    test_size();
    test_string_to_double();
    test_name_resolver();
    test_address();
    test_parse_and_print_expressions();
    test_function_name_resolution();
    test_model_context_storage();
    test_volatile_function();
    return EXIT_SUCCESS;
}
/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
